"""Converters module

This module has methods to convert data between different formats
without complicated logic. In case data cannot be converted,
`None` is returned and no exception is raised.

If data conversion requires complicated logic,
it should be in the Readers module"""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Any, Callable, Iterable, Optional, Type, TypeVar, cast

from asusrouter.tools.cleaners import clean_content

true_values = {"true", "allow", "1", "on", "enabled"}
false_values = {"false", "block", "0", "off", "disabled"}


_T = TypeVar("_T")
_E = TypeVar("_E", bound=Enum)


def clean_input(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to clean input data."""

    def wrapper(content: Any, *args, **kwargs) -> Any:
        if isinstance(content, str):
            return func(clean_string(content), *args, **kwargs)
        return func(content, *args, **kwargs)

    return wrapper


def clean_string(content: Optional[str]) -> Optional[str]:
    """Get a clean string or return None if it is empty."""

    # Not a string
    if not content or not isinstance(content, str):
        return None

    content = clean_content(content.strip())
    # Empty string
    if not content:
        return None

    return content


def flatten_dict(
    d: Optional[dict[Any, Any]],
    parent_key: str = "",
    sep: str = "_",
    exclude: Optional[str | Iterable[str]] = None,
) -> Optional[dict[str, Any]]:
    """Flatten a nested dictionary."""

    if d is None:
        return None

    if not isinstance(d, dict):
        return {}

    items = []
    exclude = (exclude,) if isinstance(exclude, str) else tuple(exclude or [])
    for k, v in d.items():
        new_key = (
            f"{parent_key}{sep}{k}" if parent_key not in ("", None) else k
        )
        # We have a dict - check it
        if isinstance(v, dict):
            # This key should be skipped
            if isinstance(new_key, str) and new_key.endswith(exclude):
                items.append((new_key, v))
                continue
            # Go recursive
            flattened = flatten_dict(v, new_key, sep, exclude)
            if flattened is not None:
                items.extend(flattened.items())
            continue
        # Not a dict - add it
        items.append((new_key, v))
    return dict(items)


def get_arguments(
    args: str | tuple[str, ...], **kwargs: Any
) -> Any | tuple[Optional[Any], ...]:
    """Get the arguments from kwargs."""

    # Make sure args is a tuple
    if not isinstance(args, tuple):
        args = (args,)

    arguments = kwargs.get("arguments", {})

    found_args: list[Optional[Any]] = []

    for arg in args:
        # Skip if not a string
        if not isinstance(arg, str):
            continue
        # Get the arg and save to found_args
        arg_value = arguments.get(arg) if arguments else kwargs.get(arg)
        found_args.append(arg_value)

    if len(found_args) == 1:
        return found_args[0]

    return tuple(found_args) if found_args else None


def get_enum_key_by_value(
    enum: Type[_E], value: Any, default: Optional[_E] = None
) -> _E:
    """Get the enum key by value"""

    if issubclass(enum, Enum):
        for enum_value in enum:
            if enum_value.value == value:
                return enum_value

    if default is not None:
        return default

    raise ValueError(f"Invalid value: {value}")


def handle_none_content(
    content: Optional[_T], default: Optional[_T]
) -> Optional[_T]:
    """Return the default value if content is None, else return the content."""

    if content is None:
        return default
    return content


def int_as_bits(value: int) -> list[bool]:
    """Convert an integer to a list of bits."""

    if not isinstance(value, int):
        return []

    # Negative values are not supported
    if value < 0:
        return []

    # Zero is a special case
    if value == 0:
        return [False]

    return [bool(value & (1 << i)) for i in range(value.bit_length())]


def int_as_capabilities(
    value: int, capabilities: Type[Enum]
) -> dict[Enum, bool]:
    """Convert an integer to a dict of capabilities."""

    # Check if the capabilities is an enum
    if not is_enum(capabilities) or not isinstance(value, int):
        return {}

    # Convert the value to a list of bits
    bits = int_as_bits(value)

    result = {}

    # For each capability in the capabilities
    # Considering key as a capability name and value as a capability bit
    for capability in capabilities:
        # Check that the capability is an integer
        if not isinstance(capability.value, int):
            continue

        # Check if the bit is set
        result[capability] = (
            bits[capability.value] if capability.value < len(bits) else False
        )

    return result


def is_enum(v) -> bool:
    """Check if the value is an enum."""

    return isinstance(v, type) and issubclass(v, Enum)


def list_from_dict(raw: Optional[dict[Any, Any] | list[Any]]) -> list[str]:
    """Return dictionary keys as list."""

    if isinstance(raw, list):
        return raw

    if not isinstance(raw, dict):
        return []

    return list(raw.keys())


def nvram_get(
    content: Optional[list[str] | str],
) -> Optional[list[tuple[str, ...]]]:
    """Convert values to NVRAM request."""

    if not content:
        return None

    if not isinstance(content, (list, str)):
        content = str(content)

    if isinstance(content, str):
        content = [content]

    return [("nvram_get", value) for value in content]


def run_method(
    value: Any, method: Optional[Callable[..., Any] | list[Callable[..., Any]]]
) -> Any:
    """Run a method or a list of methods on a value and return the result."""

    if not method:
        return value

    if not isinstance(method, list):
        method = [method]

    for func in method:
        if is_enum(func):
            try:
                value = func(value)
            except ValueError:
                value = func.UNKNOWN if hasattr(func, "UNKNOWN") else None
        else:
            value = func(value)

    return value


@clean_input
def safe_bool(content: Optional[str | int | float | bool]) -> Optional[bool]:
    """Read the content as boolean or return None."""

    if content is None:
        return None

    if isinstance(content, bool):
        return content
    if isinstance(content, (int, float)):
        return content != 0
    if isinstance(content, str):
        content = content.lower()
        if content in true_values:
            return True
        if content in false_values:
            return False

    return None


def safe_convert(
    convert_func: Callable[[str | int | float], _T],
    content: Optional[str | int | float],
    default: Optional[_T] = None,
    fallback_func: Optional[Callable[[Any], _T]] = None,
) -> Optional[_T]:
    """Try to convert the content using the conversion function,
    return the default value if it fails."""

    if content is None:
        return default
    try:
        return convert_func(content)
    except ValueError:
        if fallback_func is not None and isinstance(content, str):
            try:
                return fallback_func(content)
            except ValueError:
                pass
        return default


@clean_input
def safe_datetime(content: Optional[str]) -> Optional[datetime]:
    """Read the content as datetime or return None."""

    if not content:
        return None

    try:
        return datetime.fromisoformat(content)
    except (ValueError, TypeError):
        try:
            return datetime.strptime(content, "%a, %d %b %Y %H:%M:%S %z")
        except (ValueError, TypeError):
            return None


def safe_enum(
    enum: Type[_E],
    value: Any,
    default_value: Optional[Any] = None,
    default: Optional[_E] = None,
) -> Optional[_E]:
    """Get the enum key by value"""

    # Fast return
    # Return the default enum member
    if not default_value and not value and default is not None:
        return default

    if issubclass(enum, Enum):
        _def_enum_value = None
        # Go through the enum values
        for enum_value in enum:
            # Check for the value
            if enum_value.value == value:
                # Fast return
                return enum_value
            # Check for the default value
            if enum_value.value == default_value:
                _def_enum_value = enum_value
        # Return the default value
        if _def_enum_value is not None:
            return _def_enum_value

    return None


@clean_input
def safe_exists(content: Optional[str]) -> bool:
    """Read the content as boolean or return None."""

    if content is None:
        return False

    return True


@clean_input
def safe_float(
    content: Optional[str | int | float], default: Optional[float] = None
) -> Optional[float]:
    """Read the content as float or return None."""

    content = cast(
        Optional[str | int | float], handle_none_content(content, default)
    )
    return safe_convert(float, content, default)


@clean_input
def safe_int(
    content: Optional[str | int | float],
    default: Optional[int] = None,
    base: int = 10,
) -> Optional[int]:
    """Read the content as int or return the default value (None if not specified)."""

    content = cast(
        Optional[str | int | float], handle_none_content(content, default)
    )
    if isinstance(content, str):
        return safe_convert(
            lambda x: int(x, base=base),
            content,
            default,
            lambda x: int(float(x)),
        )
    return safe_convert(
        int, content, default if isinstance(default, int) else None
    )


def safe_list(content: Any) -> list[Any]:
    """Read any content as a list."""

    if isinstance(content, list):
        return content

    if content is None:
        return []

    return [content]


def safe_list_csv(content: Optional[str]) -> list[str]:
    """Read the list as comma separated values."""

    return safe_list_from_string(content, ",")


@clean_input
def safe_list_from_string(
    content: Optional[str], delimiter: str = " "
) -> list[str]:
    """Read the content as list or return empty list."""

    if not isinstance(content, str):
        return []

    return content.split(delimiter)


@clean_input
def safe_return(content: Any) -> Any:
    """Return the content."""

    return content


def safe_speed(
    current: (int | float),
    previous: (int | float),
    time_delta: (int | float) | None = None,
) -> float:
    """Calculate speed.

    Allows calculation only of positive speed, otherwise returns 0.0."""

    if time_delta is None or time_delta == 0.0:
        return 0.0

    diff = current - previous if current > previous else 0.0

    return diff / time_delta


def safe_time_from_delta(content: str) -> datetime:
    """Transform time delta to the date in the past."""

    return datetime.now(timezone.utc).replace(
        microsecond=0, tzinfo=timezone.utc
    ) - safe_timedelta_long(content)


@clean_input
def safe_timedelta_long(content: Optional[str]) -> timedelta:
    """Transform connection timedelta of the device to a proper
    datetime object when the device was connected"""

    if not content:
        return timedelta()

    part = content.split(":")
    try:
        return timedelta(
            hours=int(part[-3]), minutes=int(part[-2]), seconds=int(part[-1])
        )
    except (ValueError, IndexError):
        return timedelta()


def safe_unpack_key(
    content: tuple[
        str, Optional[Callable[..., Any]] | list[Callable[..., Any]]
    ]
    | str
    | tuple[str],
) -> tuple[str, Optional[Callable[..., Any] | list[Callable[..., Any]]]]:
    """
    Unpacks a tuple containing a key and a method.

    The input can be a tuple of a string and a method, a single string,
    or a single-item tuple with a string.
    If the input is a string or a single-item tuple, the returned method is None.
    If the input is a tuple of a string and a method, both are returned as is.

    Args:
        content: A tuple of a string and a method, a single string,
        or a single-item tuple with a string.

    Returns:
        A tuple containing a string and a method.
        If no method was provided in the input, None is returned as the method.
    """

    if isinstance(content, tuple):
        key = content[0]
        if len(content) > 1:
            content = cast(
                tuple[
                    str,
                    Optional[Callable[..., Any]] | list[Callable[..., Any]],
                ],
                content,
            )
            methods = content[1]
            if methods is not None and not (
                callable(methods) or isinstance(methods, Iterable)
            ):
                methods = None
        else:
            methods = None
        return key, methods

    # No method selected
    return content, None


def safe_unpack_keys(
    content: tuple[
        str, str, Optional[Callable[..., Any]] | list[Callable[..., Any]]
    ]
    | tuple[str, str]
    | str,
) -> tuple[Any, ...]:
    """Method to unpack key/key_to_use/method tuple
    even if some values are missing."""

    if isinstance(content, tuple):
        # All 3 values are present
        if len(content) == 3:
            return content

        # No method selected
        if len(content) == 2:
            return content + (None,)

    # No method and key_to_use selected
    # We need to replace key_to_use with key
    new_content = (content, content)
    return new_content + (None,)


def safe_usage(used: int | float, total: int | float) -> float:
    """Calculate usage in percents.

    Allows calculation only of positive usage, otherwise returns 0.0."""

    if total == 0:
        return 0.0

    usage = round(used / total * 100, 2)

    # Don't allow negative usage
    if usage < 0:
        return 0.0

    return usage


def safe_usage_historic(
    used: int | float,
    total: int | float,
    prev_used: int | float,
    prev_total: int | float,
) -> float:
    """Calculate usage in percents for difference between current and previous values

    This method is just an interface to calculate usage using `usage` method"""

    used_diff = used - prev_used
    total_diff = total - prev_total

    # Don't allow negative differences
    if used_diff < 0 or total_diff < 0:
        return 0.0

    return safe_usage(used_diff, total_diff)


def safe_timestamp_to_utc(value: Optional[int]) -> Optional[datetime]:
    """Convert timestamp to UTC datetime."""

    if value is None:
        return None

    try:
        return datetime.fromtimestamp(value, timezone.utc)
    except (ValueError, TypeError, OSError):
        try:
            return datetime.fromtimestamp(value / 1000, timezone.utc)
        except (ValueError, TypeError, OSError):
            return None


def safe_utc_to_timestamp(value: Optional[datetime]) -> Optional[float]:
    """Convert UTC datetime to timestamp."""

    if value is None or not isinstance(value, datetime):
        return None

    return value.timestamp()


def safe_utc_to_timestamp_milli(value: Optional[datetime]) -> Optional[int]:
    """Convert UTC datetime to timestamp in milliseconds."""

    _timestamp = safe_utc_to_timestamp(value)

    if _timestamp is None:
        return None

    return int(_timestamp * 1000)


def scale_value_int(
    value: int,
    scale: int,
    scale_from: Optional[int] = None,
) -> int:
    """Scale the value from the custom scale."""

    if scale_from is None or scale_from == scale:
        return value

    return (
        min(round(value * scale / scale_from), scale) if scale_from != 0 else 0
    )



"""Dump tools for AsusRouter.

This module contains all needed to dump the raw data from the router
for the following analysis and testing."""

from __future__ import annotations

import asyncio
import json
import os
import zipfile
from datetime import datetime, timezone
from typing import Any

from asusrouter.modules.endpoint import Endpoint


class AsusRouterDump:
    """AsusRouter Dumper."""

    _endpoint: Endpoint
    _datetime: str

    def __init__(
        self, output_folder: str, full_dump: bool = False, archive: bool = True
    ) -> None:
        """Initialize."""

        self.log: dict[str, str] = {}

        self._output_folder = output_folder
        self.full_dump = full_dump
        self.zip = archive

        self._init_datetime = datetime.now(timezone.utc)

        if self.zip:
            self._zipfile = zipfile.ZipFile(
                os.path.join(
                    self._output_folder,
                    f"AsusRouter-{self._init_datetime.isoformat().replace(':', '-')}.zip",
                ),
                "w",
            )
        else:
            # Create subfolder for the dump
            self._output_folder = os.path.join(
                output_folder,
                f"AsusRouter-{self._init_datetime.isoformat().replace(':', '-')}",
            )
            os.makedirs(self._output_folder)

    def __enter__(self):
        """Enter the runtime context related to this object."""

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context and save the log."""

        if self.zip:
            # Write the log data to the zip file
            self._zipfile.writestr("log.json", json.dumps(self.log, default=str))
            self._zipfile.close()
        else:
            log_filename = os.path.join(self._output_folder, "log.json")
            with open(log_filename, "w", encoding="utf-8") as f:
                json.dump(self.log, f, default=str)

    async def dump(
        self,
        endpoint: Endpoint,
        payload: dict[str, Any],
        resp_status: int,
        resp_headers: dict[str, Any],
        resp_content: bytes,
    ):
        """Dump the data."""

        self._endpoint = endpoint
        self._datetime = datetime.now(timezone.utc).isoformat().replace(":", "-")

        self.log[self._datetime] = f"Endpoint.{endpoint.name}"

        # Write the response content to a file
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._write_content, resp_content)

        # Write the metadata to a file
        metadata = {
            "endpoint": endpoint,
            "payload": payload,
            "resp_status": resp_status,
            "resp_headers": resp_headers,
        }
        await loop.run_in_executor(None, self._write_metadata, metadata)

    def _write_content(self, content: str):
        """Write the content to a file."""

        if self.full_dump:
            filename = f"{self._datetime}-{self._endpoint}.content"
            if self.zip:
                self._zipfile.writestr(filename, content)
            else:
                with open(
                    os.path.join(self._output_folder, filename), "w", encoding="utf-8"
                ) as f:
                    f.write(content)

    def _write_metadata(self, metadata: dict[str, Any]):
        """Write the metadata to a file."""

        if self.full_dump:
            filename = f"{self._datetime}-{self._endpoint}.json"
            if self.zip:
                self._zipfile.writestr(filename, json.dumps(metadata, default=str))
            else:
                with open(
                    os.path.join(self._output_folder, filename), "w", encoding="utf-8"
                ) as f:
                    json.dump(metadata, f, default=str)



"""Color module."""

from __future__ import annotations

import logging
from typing import Any, Optional

from asusrouter.tools.converters import clean_input, safe_int, scale_value_int

_LOGGER = logging.getLogger(__name__)

DEFAULT_COLOR = (0, 0, 0)
DEFAULT_COLOR_SCALE = 255
DEFAULT_COLOR_SCALE_ASUS = 128
DEFAULT_COLOR_DELIMITER = ","


def average_color(
    colors: ColorRGB | list[ColorRGB],
) -> ColorRGB:
    """Get the average color."""

    if isinstance(colors, ColorRGB):
        return colors

    if not colors:
        return ColorRGB()

    # Get the scale as the maximum scale from the colors
    scale = max(color.scale for color in colors)

    # Calculate the average color components
    num_colors = len(colors)
    red = sum(color.r for color in colors) // num_colors
    green = sum(color.g for color in colors) // num_colors
    blue = sum(color.b for color in colors) // num_colors

    return ColorRGB(red, green, blue, scale=scale)


@clean_input
def color_zone(
    color_str: Optional[str],
    delimiter: str = ",",
) -> int:
    """Return the number of color zones."""

    if not color_str:
        return 0

    return (len(color_str.split(delimiter))) // 3


@clean_input
def parse_colors(
    color_str: Optional[str],
    delimiter: str = ",",
    scale: int = DEFAULT_COLOR_SCALE_ASUS,
) -> Optional[list[ColorRGBB]]:
    """Parse the colors from the string."""

    if not color_str:
        return None

    # Split the string to separate groups of 3 digits
    channels = color_str.split(delimiter)
    if len(channels) % 3 != 0:
        _LOGGER.warning(
            "Invalid color string: %s, but will parse what is possible",
            color_str,
        )

    colors = []
    for i in range(0, len(channels), 3):
        color = ColorRGBB()
        if i + 2 >= len(channels):
            break
        color.from_rgbwb(
            rgb=(
                safe_int(channels[i]),
                safe_int(channels[i + 1]),
                safe_int(channels[i + 2]),
            ),
            scale=scale,
        )
        colors.append(color)

    return colors if len(colors) > 0 else None


class ColorRGB:
    """RGB color class."""

    def __init__(
        self,
        r: int | tuple[int, int, int] | Any = DEFAULT_COLOR,
        g: Optional[int] = None,
        b: Optional[int] = None,
        scale: int = DEFAULT_COLOR_SCALE_ASUS,
    ) -> None:
        """Initialize the color."""

        self._r: int
        self._g: int
        self._b: int
        self._scale = scale

        # Single value(s)
        if isinstance(r, int):
            g = g if g is not None else r
            b = b if b is not None else r
            self.from_rgb((r, g, b))
            return

        # Any other than tuple
        if not isinstance(r, tuple):
            r = DEFAULT_COLOR

        self.from_rgb(r)
        return

    def _normalize_input_rgb(
        self,
        rgb: Optional[tuple[int, ...] | str],
        delimiter: str = DEFAULT_COLOR_DELIMITER,
    ) -> tuple[int, int, int]:
        """Normalize the RGB input."""

        if rgb is None:
            return DEFAULT_COLOR

        if isinstance(rgb, str):
            rgb = tuple(safe_int(value) for value in rgb.split(delimiter))
        elif isinstance(rgb, ColorRGB):
            rgb = rgb.as_tuple()
        else:
            rgb = tuple(safe_int(value) for value in rgb)

        # Add extra values if needed or remove them
        rgb = (rgb + (0,) * 3)[:3]

        return rgb[0], rgb[1], rgb[2]

    def _normalize_scale(
        self,
        values: tuple[int, ...],
        scale: int,
    ) -> tuple[int, ...]:
        """Normalize the input values to the scale."""

        _brightness = max(values)
        if _brightness > scale:
            values = tuple(
                scale_value_int(value, scale, _brightness) for value in values
            )

        return values

    def from_rgb(
        self,
        rgb: tuple[int, ...] | str,
        delimiter: str = DEFAULT_COLOR_DELIMITER,
    ) -> None:
        """Load a color from RGB values."""

        # Prepare and normalize the input
        rgb = self._normalize_scale(
            self._normalize_input_rgb(rgb, delimiter),
            self._scale,
        )

        # Set the RGB values
        self._r, self._g, self._b = rgb

    def from_rgbs(
        self,
        rgb: tuple[int, int, int] | str,
        delimiter: str = DEFAULT_COLOR_DELIMITER,
        scale: Optional[int] = None,
    ) -> None:
        """Load a color from RGB + Scale values."""

        self._scale = scale or self._scale
        self.from_rgb(rgb, delimiter)

    def as_tuple(self) -> tuple[int, int, int]:
        """Return the color as tuple."""

        return self._r, self._g, self._b

    def __repr__(self) -> str:
        """Return the color as string."""

        return f"{self._r},{self._g},{self._b}"

    def __str__(self) -> str:
        """Return the color as string."""

        return f"{self._r},{self._g},{self._b}"

    def __eq__(self, other: object) -> bool:
        """Check if the colors are equal."""

        if not isinstance(other, ColorRGB):
            return False
        return (
            self._r == other._r
            and self._g == other._g
            and self._b == other._b
            and self._scale == other._scale
        )

    def __hash__(self) -> int:
        """Return the hash of the object."""

        return hash((self._r, self._g, self._b, self._scale))

    @property
    def r(self) -> int:
        return self._r

    @property
    def g(self) -> int:
        return self._g

    @property
    def b(self) -> int:
        return self._b

    @property
    def scale(self) -> int:
        return self._scale

    @property
    def color_brightness(self) -> int:
        return max(self._r, self._g, self._b)


class ColorRGBB(ColorRGB):
    """RGB + Brightness color class."""

    def __init__(
        self,
        rgb: tuple[int, int, int] | ColorRGB = DEFAULT_COLOR,
        br: Optional[int] = None,
        scale: int = DEFAULT_COLOR_SCALE_ASUS,
    ) -> None:
        """Initialize the color."""

        self._scale = scale
        self.from_rgbwb(rgb)
        self._br = min(br, self._scale) if br is not None else self._scale

    def set_brightness(self, br: int) -> None:
        """Set the brightness."""

        self._br = max(0, min(br, self._scale))

    def _from_rgb(
        self,
        rgb: tuple[int, ...] | str,
        delimiter: str = DEFAULT_COLOR_DELIMITER,
        scale: Optional[int] = None,
    ) -> tuple[int, int, int]:
        """Read the RGB values."""

        # Prepare and normalize the input
        rgb = self._normalize_scale(
            self._normalize_input_rgb(rgb, delimiter),
            scale or self._scale,
        )

        return rgb[0], rgb[1], rgb[2]

    def from_rgb(
        self,
        rgb: tuple[int, ...] | str | ColorRGB,
        delimiter: str = DEFAULT_COLOR_DELIMITER,
        scale: Optional[int] = None,
    ) -> None:
        """Load a color from RGB values."""

        if isinstance(rgb, ColorRGB):
            rgb = (rgb.r, rgb.g, rgb.b)

        self._scale = scale or self._scale
        rgb = self._from_rgb(rgb, delimiter, self._scale)

        self._r, self._g, self._b = rgb

    def from_rgbwb(
        self,
        rgb: tuple[int, ...] | ColorRGB,
        delimiter: str = DEFAULT_COLOR_DELIMITER,
        scale: Optional[int] = None,
    ) -> None:
        """Load a color from RGB/wb (RGB with B embedded) values."""

        if isinstance(rgb, ColorRGB):
            rgb = (rgb.r, rgb.g, rgb.b)

        self._scale = scale or self._scale
        rgb = self._from_rgb(rgb, delimiter, self._scale)

        # Get color brightness
        self._br = max(rgb)

        # Rescale RGB to the full scale
        rgb = tuple(
            scale_value_int(value, self._scale, self._br) for value in rgb
        )
        self._r, self._g, self._b = rgb

    def _to_rgb(
        self,
        rgbb: Optional[tuple[int, ...]] = None,
    ) -> tuple[int, int, int]:
        """Return RGBB color as RGB."""

        if rgbb is not None and len(rgbb) >= 4:
            return (
                scale_value_int(rgbb[0], rgbb[3], self._scale),
                scale_value_int(rgbb[1], rgbb[3], self._scale),
                scale_value_int(rgbb[2], rgbb[3], self._scale),
            )

        return (
            scale_value_int(self._r, self._br, self._scale),
            scale_value_int(self._g, self._br, self._scale),
            scale_value_int(self._b, self._br, self._scale),
        )

    def to_rgb(
        self,
        scale: Optional[int] = None,
    ) -> ColorRGB:
        """Return RGBB color as RGB."""

        rgb = self._to_rgb()

        if scale is not None:
            return ColorRGB(self._to_rgb((rgb[0], rgb[1], rgb[2], scale)))

        return ColorRGB(rgb, scale=self._scale)

    def __repr__(self) -> str:
        """Return the color as string."""

        return f"{self._r},{self._g},{self._b},{self._br},{self._scale}"

    def __str__(self) -> str:
        """Return the color as string."""

        return f"{self._r},{self._g},{self._b},{self._br},{self._scale}"

    @property
    def brightness(self) -> int:
        return self._br



"""Reading tools for AsusRouter"""

from __future__ import annotations

import json
import logging
import re
from typing import Any, Optional

from asusrouter.const import ContentType
from asusrouter.tools.converters import clean_input

_LOGGER = logging.getLogger(__name__)


# Random symbols to avoid json errors
RANDOM_SYMBOLS: list[str] = [
    "\u0000",
    "\u0001",
    "\u0002",
    "\u0003",
    "\u0004",
    "\u0005",
    "\u0006",
    "\u0007",
    "\u0008",
    "\u0009",
]


def merge_dicts(
    data: dict[Any, Any], merge_data: dict[Any, Any]
) -> dict[Any, Any]:
    """This methods merges two nested dicts into a single one
    while keeping all the existing values"""

    # Create a new dictionary to store the merged data
    merged_data = data.copy()

    # Go through the merge data and fill the merged data
    for key, value in merge_data.items():
        # Check if the value is a dict
        if isinstance(value, dict):
            # Check if the key is in the merged data
            if key not in merged_data or merged_data[key] is None:
                # Add the key to the merged data
                merged_data[key] = {}
            # Merge the dicts
            merged_data[key] = merge_dicts(merged_data[key], value)
            # Continue to the next value
            continue
        # Check if the key is in the merged data
        if key not in merged_data:
            # Add the key to the merged data
            merged_data[key] = value
        else:
            # If the key is already in the merged data, compare the values
            if merged_data[key] is None:
                # If the value in the merged data is None, take the value from the merge data
                merged_data[key] = value
            elif value is None:
                # If the value in the merge data is not None
                # and the value in the merge data is None, keep the value in the merged data
                pass
            else:
                # If both values are not None, keep the value from the merged data
                pass

    # Return the merged data
    return merged_data


def read_as_snake_case(data: str) -> str:
    """Convert a string to snake case"""

    string = (
        re.sub(r"(?<=[a-z])(?=[A-Z])|[^a-zA-Z]", " ", data)
        .strip()
        .replace(" ", "_")
    )
    result = "".join(string.lower())
    while "__" in result:
        result = result.replace("__", "_")

    return result


def read_content_type(headers: dict[str, str]) -> ContentType:
    """Get the content type from the headers"""

    # Get the content type from the headers
    content_type = headers.get("content-type", "").split(";")[0].strip()
    # Find the content type in ContentType enum and return correct ContentType enum
    for content_type_enum in ContentType:
        if content_type_enum.value == content_type:
            return content_type_enum

    # If the content type is not found, return the content type as text
    return ContentType.UNKNOWN


@clean_input
def read_js_variables(content: str) -> dict[str, Any]:
    """Get all the JS variables from the content"""

    # Create a dict to store the data
    js_variables: dict[str, Any] = {}

    # Split the content into lines, strip them, remove the last semicolon
    lines = content.splitlines()
    lines = [line.strip() for line in lines]
    lines = [line[:-1] if line.endswith(";") else line for line in lines]

    # Create a regex to match the data. Consider the following:
    # The key is a string which can contain letters, numbers, underscores
    # The key and the value are separated by an equal sign (surrounded by spaces or not)
    regex = re.compile(r"(\w+)\s*=\s*(.*)")

    # Go through the lines and fill the match data to the dict
    for line in lines:
        match = regex.match(line)
        if match:
            key, value = match.groups()
            # Clean the value of quotes if it starts and ends with them
            if (value.startswith("'") and value.endswith("'")) or (
                value.startswith('"') and value.endswith('"')
            ):
                value = value[1:-1]
            js_variables[key] = value

    # Return the JS variables
    return js_variables


@clean_input
def read_json_content(content: Optional[str]) -> dict[str, Any]:
    """Get the json content"""

    if not content:
        return {}

    # Random control characters to avoid json errors
    for symbol in RANDOM_SYMBOLS:
        content = content.replace(symbol, "")

    # Handle missing values in JSON
    content = re.sub(r"\s*,\s*,", ", ", content)
    content = re.sub(r"^\s*{\s*,", "{", content)
    content = re.sub(r",\s*}\s*$", "}", content)

    # Handle keys without values
    content = re.sub(r":\s*(,|\})", ": null\\1", content)

    # Return the json content
    try:
        return json.loads(content.encode().decode("utf-8-sig"))
    except json.JSONDecodeError as ex:
        _LOGGER.error(
            "Unable to decode json content with exception `%s`.\
                Please, copy this and fill in a bug report: %s",
            ex,
            content,
        )
        return {}


@clean_input
def readable_mac(raw: Optional[str]) -> bool:
    """Checks if string is MAC address"""

    if isinstance(raw, str):
        if re.search(
            re.compile("^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"), raw
        ):
            return True

    return False



"""WAN endpoint module."""

from __future__ import annotations

from enum import Enum


class AsusDualWAN(str, Enum):
    """Dual WAN class."""

    FAILOVER = "fo"
    FALLBACK = "fb"
    LOAD_BALANCE = "lb"



"""Error module."""

from __future__ import annotations

from typing import Any, Optional

import aiohttp


class AsusRouterError(Exception):
    """Base class for errors in AsusRouter library."""

    def __init__(
        self, *args: Any, message: Optional[str] = None, **_kwargs: Any
    ) -> None:
        """Initialise base error."""

        super().__init__(*args, message)


class AsusRouterDataError(AsusRouterError):
    """Any error with received data."""


class AsusRouterConnectionError(AsusRouterError, aiohttp.ClientConnectionError):
    """Connection error."""


class AsusRouterTimeoutError(AsusRouterError):
    """Timeout error."""


class AsusRouterSessionError(AsusRouterError):
    """Session error."""


class AsusRouter404Error(AsusRouterError):
    """Page not found error."""


class AsusRouterIdentityError(AsusRouterError):
    """Identity error."""


class AsusRouterAccessError(AsusRouterError):
    """Access error."""


class AsusRouterLogoutError(AsusRouterError):
    """Logout error."""


class AsusRouterServiceError(AsusRouterError):
    """Service error."""
